"""Methods for abstracting SQL selections and parsing results."""

from abc import ABC
from abc import abstractmethod

from charli3_dendrite.dataclasses.models import DendriteBaseModel
from charli3_dendrite.dataclasses.models import PoolStateList
from charli3_dendrite.dataclasses.models import ScriptReference
from charli3_dendrite.dataclasses.models import SwapTransactionList


class AbstractDBSyncStructure(ABC):
    """Abstract class with required methods for SQL queries."""

    @classmethod
    @abstractmethod
    def select(cls) -> str:
        """The selectin part of a DBSync query."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def parse(cls, data: dict | list[dict]) -> DendriteBaseModel:
        """Parse data returned from a dbsync query."""
        raise NotImplementedError


class PoolSelector(AbstractDBSyncStructure):
    """Query selections and parsing classes for AMM pools."""

    @classmethod
    def select(cls) -> str:
        """Select SQL query for swap pools."""
        return """
SELECT txo.address,
ENCODE(tx.hash, 'hex') as "tx_hash",
txo.index as "tx_index",
EXTRACT(
	epoch
	FROM block.time
)::INTEGER AS "block_time",
tx.block_index as "block_index",
ENCODE(block.hash,'hex') as "block_hash",
ENCODE(datum.hash,'hex') as "datum_hash",
ENCODE(datum.bytes,'hex') as "datum_cbor",
COALESCE (
    json_build_object('lovelace',txo.value::TEXT)::jsonb || (
        SELECT json_agg(
            json_build_object(
                CONCAT(encode(ma.policy, 'hex'), encode(ma.name, 'hex')),
                mto.quantity::TEXT
            )
        )
        FROM ma_tx_out mto
        JOIN multi_asset ma ON (mto.ident = ma.id)
        WHERE mto.tx_out_id = txo.id
    )::jsonb,
    jsonb_build_array(json_build_object('lovelace',txo.value::TEXT)::jsonb)
) AS "assets",
(txo.inline_datum_id IS NOT NULL OR txo.reference_script_id IS NOT NULL) as "plutus_v2"
"""

    @classmethod
    def parse(cls, data: dict | list[dict]) -> PoolStateList:
        """Parse pools from a query."""
        return PoolStateList.model_validate(data)


class UTxOSelector(AbstractDBSyncStructure):
    """Selection queries and parsing classes for reference scripts and datums."""

    @classmethod
    def select(cls) -> str:
        """Selection SQL query for reference scripts and datums."""
        return """
SELECT ENCODE(tx.hash, 'hex') as "tx_hash",
tx_out.index as "tx_index",
address.address,
ENCODE(datum.hash,'hex') as "datum_hash",
ENCODE(datum.bytes,'hex') as "datum_cbor",
COALESCE (
    json_build_object('lovelace',tx_out.value::TEXT)::jsonb || (
        SELECT json_agg(
            json_build_object(
                CONCAT(encode(ma.policy, 'hex'), encode(ma.name, 'hex')),
                mto.quantity::TEXT
            )
        )
        FROM ma_tx_out mto
        JOIN multi_asset ma ON (mto.ident = ma.id)
        WHERE mto.tx_out_id = tx_out.id
    )::jsonb,
    jsonb_build_array(json_build_object('lovelace',tx_out.value::TEXT)::jsonb)
) AS "assets",
ENCODE(s.bytes, 'hex') as "script" """

    @classmethod
    def parse(cls, data: dict | list[dict]) -> ScriptReference:
        """Parsing class for UTxOs containing a reference script or datum."""
        return ScriptReference.model_validate(data)


class OrderSelector(AbstractDBSyncStructure):
    """SQL query for orders and the associated pydantic parsing class."""

    @classmethod
    def select(cls) -> str:
        """The SQL select statement for orders."""
        return """
SELECT (
	SELECT array_agg(DISTINCT add.address)
	FROM tx_out txo
	LEFT JOIN address add ON add.id = txo.address_id
	WHERE txo.consumed_by_tx_id = tx.id
) AS "submit_address_inputs",
address.address as "submit_address_stake",
ENCODE(tx.hash, 'hex') as "submit_tx_hash",
txo_stake.index as "submit_tx_index",
ENCODE(block.hash,'hex') as "submit_block_hash",
EXTRACT(
	epoch
	FROM block.time
)::INTEGER AS "submit_block_time",
tx.block_index AS "submit_block_index",
(
	SELECT array_agg(tx_metadata.json)
	FROM tx_metadata
	WHERE tx.id = tx_metadata.tx_id
) AS "submit_metadata",
COALESCE(
	json_build_object('lovelace',txo_stake.value::TEXT)::jsonb || (
		SELECT json_agg(
			json_build_object(
				CONCAT(encode(ma.policy, 'hex'), encode(ma.name, 'hex')),
				mto.quantity::TEXT
			)
		)
		FROM ma_tx_out mto
		JOIN multi_asset ma ON (mto.ident = ma.id)
		WHERE mto.tx_out_id = txo_stake.id
	)::jsonb,
	jsonb_build_array(json_build_object('lovelace',txo_stake.value::TEXT)::jsonb)
) AS "submit_assets",
ENCODE(datum.hash,'hex') as "submit_datum_hash",
ENCODE(datum.bytes,'hex') as "submit_datum_cbor",
txo_output.address,
ENCODE(txo_output.tx_hash, 'hex') as "tx_hash",
txo_output.tx_index as "tx_index",
EXTRACT(
	epoch
	FROM txo_output.block_time
)::INTEGER AS "block_time",
txo_output.block_index AS "block_index",
ENCODE(txo_output.block_hash,'hex') AS "block_hash",
ENCODE(txo_output.datum_hash, 'hex') AS "datum_hash",
ENCODE(txo_output.datum_bytes, 'hex') AS "datum_cbor",
COALESCE(
	json_build_object('lovelace',txo_output.value::TEXT)::jsonb || (
		SELECT json_agg(
			json_build_object(
				CONCAT(encode(ma.policy, 'hex'), encode(ma.name, 'hex')),
				mto.quantity::TEXT
			)
		)
		FROM ma_tx_out mto
		JOIN multi_asset ma ON (mto.ident = ma.id)
		WHERE mto.tx_out_id = txo_output.tx_id
	)::jsonb,
	jsonb_build_array(json_build_object('lovelace',txo_output.value::TEXT)::jsonb)
) AS "assets",
(
    txo_output.inline_datum_id IS NOT NULL OR txo_output.reference_script_id IS NOT NULL
) as "plutus_v2"
"""

    @classmethod
    def parse(cls, data: dict | list[dict]) -> SwapTransactionList:
        """Parse and validate orders."""
        return SwapTransactionList.model_validate(data)



"""WingRiders DEX Module."""

from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from typing import ClassVar
from typing import Union

from pycardano import Address
from pycardano import PlutusData
from pycardano import PlutusV1Script
from pycardano import PlutusV2Script
from pycardano.plutus import RawDatum

from charli3_dendrite.dataclasses.datums import AssetClass
from charli3_dendrite.dataclasses.datums import OrderDatum
from charli3_dendrite.dataclasses.datums import PlutusFullAddress
from charli3_dendrite.dataclasses.datums import PoolDatum
from charli3_dendrite.dataclasses.models import Assets
from charli3_dendrite.dataclasses.models import OrderType
from charli3_dendrite.dataclasses.models import PoolSelector
from charli3_dendrite.dexs.amm.amm_types import AbstractConstantProductPoolState
from charli3_dendrite.dexs.amm.amm_types import AbstractStableSwapPoolState
from charli3_dendrite.dexs.core.errors import NotAPoolError


@dataclass
class WingriderAssetClass(PlutusData):
    """Encode a pair of assets for the WingRiders DEX."""

    CONSTR_ID = 0

    asset_a: AssetClass
    asset_b: AssetClass

    @classmethod
    def from_assets(cls, in_assets: Assets, out_assets: Assets):
        """Create a WingRiderAssetClass from a pair of assets."""
        merged = in_assets + out_assets
        if in_assets.unit() == merged.unit():
            return cls(
                asset_a=AssetClass.from_assets(in_assets),
                asset_b=AssetClass.from_assets(out_assets),
            )
        else:
            return cls(
                asset_a=AssetClass.from_assets(out_assets),
                asset_b=AssetClass.from_assets(in_assets),
            )


@dataclass
class RewardPlutusPartAddress(PlutusData):
    """Encode a plutus address part (i.e. payment, stake, etc)."""

    CONSTR_ID = 1
    address: bytes


@dataclass
class RewardPlutusFullAddress(PlutusFullAddress):
    """A full address, including payment and staking keys."""

    CONSTR_ID = 0

    payment: RewardPlutusPartAddress


@dataclass
class WingRiderOrderConfig(PlutusData):
    """Configuration for a WingRiders order."""

    CONSTR_ID = 0

    full_address: Union[PlutusFullAddress, RewardPlutusFullAddress]
    address: bytes
    expiration: int
    assets: WingriderAssetClass

    @classmethod
    def create_config(
        cls,
        address: Address,
        expiration: int,
        in_assets: Assets,
        out_assets: Assets,
    ):
        """Create a WingRiders order configuration."""
        plutus_address = PlutusFullAddress.from_address(address)
        assets = WingriderAssetClass.from_assets(
            in_assets=in_assets,
            out_assets=out_assets,
        )

        return cls(
            full_address=plutus_address,
            address=bytes.fromhex(str(address.payment_part)),
            expiration=expiration,
            assets=assets,
        )


@dataclass
class AtoB(PlutusData):
    """A to B."""

    CONSTR_ID = 0


@dataclass
class BtoA(PlutusData):
    """B to A."""

    CONSTR_ID = 1


@dataclass
class WingRidersOrderDetail(PlutusData):
    """WingRiders order detail."""

    CONSTR_ID = 0

    direction: Union[AtoB, BtoA]
    min_receive: int

    @classmethod
    def from_assets(cls, in_assets: Assets, out_assets: Assets):
        """Create a WingRidersOrderDetail from a pair of assets."""
        merged = in_assets + out_assets
        if in_assets.unit() == merged.unit():
            return cls(direction=AtoB(), min_receive=out_assets.quantity())
        else:
            return cls(direction=BtoA(), min_receive=out_assets.quantity())


@dataclass
class WingRidersDepositDetail(PlutusData):
    """WingRiders deposit detail."""

    CONSTR_ID = 1

    min_lp_receive: int


@dataclass
class WingRidersWithdrawDetail(PlutusData):
    """WingRiders withdraw detail."""

    CONSTR_ID = 2

    min_amount_a: int
    min_amount_b: int


@dataclass
class WingRidersMaybeFeeClaimDetail(PlutusData):
    """WingRiders maybe fee claim detail."""

    CONSTR_ID = 3


@dataclass
class WingRidersStakeRewardDetail(PlutusData):
    """WingRiders stake reward detail."""

    CONSTR_ID = 4


@dataclass
class WingRidersOrderDatum(OrderDatum):
    """WingRiders order datum."""

    config: WingRiderOrderConfig
    detail: Union[
        WingRidersDepositDetail,
        WingRidersMaybeFeeClaimDetail,
        WingRidersStakeRewardDetail,
        WingRidersOrderDetail,
        WingRidersWithdrawDetail,
    ]

    @classmethod
    def create_datum(
        cls,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        batcher_fee: Assets,
        deposit: Assets,
        address_target: Address | None = None,
        datum_target: PlutusData | None = None,
    ):
        """Create a WingRiders order datum."""
        timeout = int(((datetime.utcnow() + timedelta(days=360)).timestamp()) * 1000)

        config = WingRiderOrderConfig.create_config(
            address=address_source,
            expiration=timeout,
            in_assets=in_assets,
            out_assets=out_assets,
        )
        detail = WingRidersOrderDetail.from_assets(
            in_assets=in_assets,
            out_assets=out_assets,
        )

        return cls(config=config, detail=detail)

    def address_source(self) -> Address:
        return self.config.full_address.to_address()

    def requested_amount(self) -> Assets:
        if isinstance(self.detail, WingRidersDepositDetail):
            return Assets({"lp": self.detail.min_lp_receive})
        elif isinstance(self.detail, WingRidersOrderDetail):
            if isinstance(self.detail.direction, BtoA):
                return Assets(
                    {self.config.assets.asset_a.assets.unit(): self.detail.min_receive},
                )
            else:
                return Assets(
                    {self.config.assets.asset_b.assets.unit(): self.detail.min_receive},
                )
        elif isinstance(self.detail, WingRidersWithdrawDetail):
            return Assets(
                {
                    self.config.assets.asset_a.assets.unit(): self.detail.min_amount_a,
                    self.config.assets.asset_b.assets.unit(): self.detail.min_amount_b,
                },
            )
        elif isinstance(self.detail, WingRidersMaybeFeeClaimDetail):
            return Assets({})

    def order_type(self) -> OrderType | None:
        order_type = None
        if isinstance(self.detail, WingRidersOrderDetail):
            order_type = OrderType.swap
        elif isinstance(self.detail, WingRidersDepositDetail):
            order_type = OrderType.deposit
        elif isinstance(self.detail, WingRidersWithdrawDetail):
            order_type = OrderType.withdraw

        return order_type


@dataclass
class NoDatum(PlutusData):
    """WingRiders no datum."""

    CONSTR_ID = 0


@dataclass
class HashDatum(PlutusData):
    """WingRiders hash datum."""

    CONSTR_ID = 1


@dataclass
class InlineDatum(PlutusData):
    """WingRiders inline datum."""

    CONSTR_ID = 2


@dataclass
class SwapAction(PlutusData):
    """Swap."""

    CONSTR_ID = 0

    swap_direction: Union[AtoB, BtoA]
    minimum_receive: int


@dataclass
class AddLiquidityAction(PlutusData):
    """Add Liquidity."""

    CONSTR_ID = 1

    minimum_receive: int


@dataclass
class WithdrawLiquidityAction(PlutusData):
    """Add Liquidity."""

    CONSTR_ID = 2

    minimum_receive_a: int
    minimum_receive_b: int


@dataclass
class WithdrawProjectAction(PlutusData):
    """Unknown action."""

    CONSTR_ID = 3


@dataclass
class WithdrawProtocolAction(PlutusData):
    """Unknown action."""

    CONSTR_ID = 4


@dataclass
class WingRidersV2OrderDatum(OrderDatum):
    """WingRiders order datum."""

    oil: int
    beneficiary: PlutusFullAddress
    owner_address: PlutusFullAddress
    compensation_datum: Union[bytes, list, RawDatum]
    compensation_datum_type: Union[NoDatum, HashDatum, InlineDatum]
    deadline: int
    asset_a_symbol: bytes
    asset_a_token: bytes
    asset_b_symbol: bytes
    asset_b_token: bytes
    action: Union[
        SwapAction,
        AddLiquidityAction,
        WithdrawLiquidityAction,
        WithdrawProjectAction,
        WithdrawProtocolAction,
    ]
    a_scale: int
    b_scale: int

    @classmethod
    def create_datum(
        cls,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        batcher_fee: Assets,
        deposit: Assets,
        address_target: Address | None = None,
        datum_target: PlutusData | None = None,
        a_scale: int = 1,
        b_scale: int = 1,
    ):
        """Create a WingRiders order datum."""
        timeout = int(((datetime.utcnow() + timedelta(days=360)).timestamp()) * 1000)

        merged = in_assets + out_assets
        if in_assets.unit() == merged.unit():
            direction = AtoB()
        else:
            direction = BtoA()

        plutus_address = PlutusFullAddress.from_address(address_source)

        return WingRidersV2OrderDatum(
            oil=2000000,
            beneficiary=plutus_address,
            owner_address=plutus_address,
            compensation_datum=b"",
            compensation_datum_type=NoDatum(),
            deadline=timeout,
            asset_a_symbol=(
                bytes.fromhex(merged.unit()[:56])
                if merged.unit() != "lovelace"
                else b""
            ),
            asset_a_token=(
                bytes.fromhex(merged.unit()[56:])
                if merged.unit() != "lovelace"
                else b""
            ),
            asset_b_symbol=bytes.fromhex(merged.unit(1)[:56]),
            asset_b_token=bytes.fromhex(merged.unit(1)[56:]),
            action=SwapAction(
                swap_direction=direction,
                minimum_receive=out_assets.quantity(),
            ),
            a_scale=a_scale,
            b_scale=b_scale,
        )

    def address_source(self) -> Address:
        return self.owner_address.to_address()

    def requested_amount(self) -> Assets:
        if isinstance(self.action, AddLiquidityAction):
            return Assets({"lp": self.action.minimum_receive})
        elif isinstance(self.action, SwapAction):
            if isinstance(self.action.swap_direction, BtoA):
                if self.asset_a_symbol.hex() == "":
                    asset_a = "lovelace"
                else:
                    asset_a = (self.asset_a_symbol + self.asset_a_token).hex()
                return Assets(
                    root={asset_a: self.action.minimum_receive},
                )
            else:
                asset_b = (self.asset_b_symbol + self.asset_b_token).hex()
                return Assets(
                    root={asset_b: self.action.minimum_receive},
                )
        elif isinstance(self.action, WithdrawLiquidityAction):
            return Assets(
                {
                    self.config.assets.asset_a.assets.unit(): self.action.min_amount_a,
                    self.config.assets.asset_b.assets.unit(): self.action.min_amount_b,
                },
            )

    def order_type(self) -> OrderType | None:
        order_type = None
        if isinstance(self.action, SwapAction):
            order_type = OrderType.swap
        elif isinstance(self.action, AddLiquidityAction):
            order_type = OrderType.deposit
        elif isinstance(self.action, WithdrawLiquidityAction):
            order_type = OrderType.withdraw

        return order_type


@dataclass
class LiquidityPoolAssets(PlutusData):
    """Encode a pair of assets for the WingRiders DEX."""

    CONSTR_ID = 0
    asset_a: AssetClass
    asset_b: AssetClass


@dataclass
class LiquidityPool(PlutusData):
    """Encode a liquidity pool for the WingRiders DEX."""

    CONSTR_ID = 0
    assets: LiquidityPoolAssets
    last_swap: int
    quantity_a: int
    quantity_b: int


@dataclass
class WingRidersPoolDatum(PoolDatum):
    """WingRiders pool datum."""

    lp_hash: bytes
    datum: LiquidityPool

    def pool_pair(self) -> Assets | None:
        return self.datum.assets.asset_a.assets + self.datum.assets.asset_b.assets


@dataclass
class CPPDatum(PlutusData):
    """WingRiders pool datum."""

    CONSTR_ID = 0


@dataclass
class SSPDatum(PlutusData):
    """WingRiders pool datum."""

    CONSTR_ID = 0

    parameter_d: int
    scale_a: int
    scale_b: int


@dataclass
class WingRidersV2PoolDatum(PoolDatum):
    """WingRiders pool datum."""

    CONSTR_ID = 0
    request_validator_hash: bytes
    asset_a_symbol: bytes
    asset_a_token: bytes
    asset_b_symbol: bytes
    asset_b_token: bytes
    swap_fee_in_basis: int
    protocol_fee_in_basis: int
    project_fee_in_basis: int
    reserve_fee_in_basis: int
    fee_basis: int
    agent_fee_ada: int
    last_interaction: int
    treasury_a: int
    treasury_b: int
    project_treasury_a: int
    project_treasury_b: int
    reserve_treasury_a: int
    reserve_treasury_b: int
    project_beneficiary: RawDatum
    reserve_beneficiary: RawDatum
    pool_specifics: CPPDatum

    def pool_pair(self) -> Assets | None:
        asset_a = (self.asset_a_symbol + self.asset_a_token).hex()
        if asset_a == "":
            asset_a = "lovelace"
        asset_b = (self.asset_b_symbol + self.asset_b_token).hex()
        return Assets(root={asset_a: 0, asset_b: 0})


@dataclass
class WingRidersV2SSPoolDatum(WingRidersV2PoolDatum):
    """WingRiders pool datum."""

    pool_specifics: SSPDatum


class WingRidersCPPState(AbstractConstantProductPoolState):
    """WingRiders CPP state."""

    fee: int = 35
    _batcher = Assets(lovelace=2000000)
    _deposit = Assets(lovelace=2000000)
    _stake_address: ClassVar[Address] = Address.from_primitive(
        "addr1wxr2a8htmzuhj39y2gq7ftkpxv98y2g67tg8zezthgq4jkg0a4ul4",
    )

    @classmethod
    def dex(cls) -> str:
        return "WingRiders"

    @classmethod
    def order_selector(self) -> list[str]:
        return [self._stake_address.encode()]

    @classmethod
    def pool_selector(cls) -> PoolSelector:
        return PoolSelector(
            addresses=[
                "addr1w8nvjzjeydcn4atcd93aac8allvrpjn7pjr2qsweukpnayghhwcpj",
            ],
            assets=cls.dex_policy(),
        )

    @property
    def swap_forward(self) -> bool:
        return False

    @property
    def stake_address(self) -> Address:
        return self._stake_address

    @classmethod
    def order_datum_class(self) -> type[WingRidersOrderDatum]:
        return WingRidersOrderDatum

    @classmethod
    def pool_datum_class(self) -> type[WingRidersPoolDatum]:
        return WingRidersPoolDatum

    @classmethod
    def pool_policy(cls) -> str:
        return ["026a18d04a0c642759bb3d83b12e3344894e5c1c7b2aeb1a2113a570"]

    @classmethod
    def dex_policy(cls) -> str:
        return ["026a18d04a0c642759bb3d83b12e3344894e5c1c7b2aeb1a2113a5704c"]

    @property
    def pool_id(self) -> str:
        """A unique identifier for the pool."""
        return self.pool_nft.unit()

    @classmethod
    def skip_init(cls, values) -> bool:
        if "pool_nft" in values and "dex_nft" in values:
            if cls.dex_policy()[0] not in values["dex_nft"]:
                raise NotAPoolError("Invalid DEX NFT")
            if len(values["assets"]) == 3:
                # Send the ADA token to the end
                if isinstance(values["assets"], Assets):
                    values["assets"].root["lovelace"] = values["assets"].root.pop(
                        "lovelace",
                    )
                else:
                    values["assets"]["lovelace"] = values["assets"].pop("lovelace")
            values["assets"] = Assets.model_validate(values["assets"])
            return True
        else:
            return False

    @classmethod
    def post_init(cls, values):
        super().post_init(values)

        assets = values["assets"]
        datum = WingRidersPoolDatum.from_cbor(values["datum_cbor"])

        if len(assets) == 2:
            assets.root[assets.unit(0)] -= 3000000

        assets.root[assets.unit(0)] -= datum.datum.quantity_a
        assets.root[assets.unit(1)] -= datum.datum.quantity_b

    def deposit(
        self,
        in_assets: Assets | None = None,
        out_assets: Assets | None = None,
    ):
        merged_assets = in_assets + out_assets
        if "lovelace" in merged_assets:
            return Assets(lovelace=4000000) - self.batcher_fee(
                in_assets=in_assets,
                out_assets=out_assets,
            )
        else:
            return self._deposit

    def batcher_fee(
        self,
        in_assets: Assets | None = None,
        out_assets: Assets | None = None,
        extra_assets: Assets | None = None,
    ):
        merged_assets = in_assets + out_assets
        if "lovelace" in merged_assets:
            if merged_assets["lovelace"] <= 250000000:
                return Assets(lovelace=850000)
            elif merged_assets["lovelace"] <= 500000000:
                return Assets(lovelace=1500000)
        return self._batcher


class WingRidersSSPState(AbstractStableSwapPoolState, WingRidersCPPState):
    """WingRiders SSP state."""

    fee: int = 6
    _batcher = Assets(lovelace=1500000)
    _deposit = Assets(lovelace=2000000)
    _stake_address = Address.from_primitive(
        "addr1w8z7qwzszt2lqy93m3atg2axx22yq5k7yvs9rmrvuwlawts2wzadz",
    )

    @classmethod
    def pool_selector(cls) -> PoolSelector:
        return PoolSelector(
            addresses=[
                "addr1wxvx34v0hlxzk9x0clv7as9hvhn7dlzwj5xfcf6g4n5uucg4tkd7w",
            ],
            assets=cls.dex_policy(),
        )

    @classmethod
    def pool_policy(cls) -> str:
        return ["980e8c567670d34d4ec13a0c3b6de6199f260ae5dc9dc9e867bc5c93"]

    @classmethod
    def dex_policy(cls) -> str:
        return ["980e8c567670d34d4ec13a0c3b6de6199f260ae5dc9dc9e867bc5c934c"]


class WingRidersV2CPPState(AbstractConstantProductPoolState):
    """WingRiders CPP state."""

    fee: int = 35
    _batcher = Assets(lovelace=2000000)
    _deposit = Assets(lovelace=2000000)
    _stake_address: ClassVar[Address] = Address.from_primitive(
        "addr1w8qnfkpe5e99m7umz4vxnmelxs5qw5dxytmfjk964rla98q605wte",
    )
    _pool_datum_parsed: WingRidersV2PoolDatum | None = None

    @classmethod
    def dex(cls) -> str:
        return "WingRidersV2"

    @classmethod
    def order_selector(self) -> list[str]:
        return [self._stake_address.encode()]

    @classmethod
    def pool_selector(cls) -> PoolSelector:
        return PoolSelector(
            addresses=["addr1wxhew7fmsup08qvhdnkg8ccra88pw7q5trrncja3dlszhqczc0qfe"],
            assets=cls.dex_policy(),
        )

    @property
    def swap_forward(self) -> bool:
        return False

    @property
    def pool_id(self) -> str:
        """A unique identifier for the pool."""
        return self.pool_nft.unit()

    @property
    def stake_address(self) -> Address:
        return self._stake_address

    @classmethod
    def order_datum_class(cls) -> type[WingRidersV2OrderDatum]:
        return WingRidersV2OrderDatum

    @classmethod
    def pool_datum_class(cls) -> type[WingRidersV2PoolDatum]:
        return WingRidersV2PoolDatum

    @classmethod
    def pool_policy(cls) -> list[str]:
        return ["6fdc63a1d71dc2c65502b79baae7fb543185702b12c3c5fb639ed737"]

    @classmethod
    def dex_policy(cls) -> list[str]:
        return ["6fdc63a1d71dc2c65502b79baae7fb543185702b12c3c5fb639ed7374c"]

    @classmethod
    def default_script_class(cls) -> type[PlutusV1Script] | type[PlutusV2Script]:
        """Returns default script class of type PlutusV1Script or PlutusV2Script."""
        return PlutusV2Script

    @classmethod
    def skip_init(cls, values) -> bool:
        if "pool_nft" in values and "dex_nft" in values:
            if cls.dex_policy()[0] not in values["dex_nft"]:
                raise NotAPoolError("Invalid DEX NFT")

            if not isinstance(values["assets"], Assets):
                values["assets"] = Assets.model_validate(values["assets"])

            cls._post_init(values)

            return True
        else:
            return False

    @property
    def pool_datum(self) -> PlutusData:
        """The pool state datum."""
        if self._pool_datum_parsed is None:
            self._pool_datum_parsed = self.pool_datum_class().from_cbor(self.datum_cbor)
        return self._pool_datum_parsed

    @classmethod
    def _post_init(cls, values):
        super().post_init(values)

        try:
            datum = WingRidersV2SSPoolDatum.from_cbor(values["datum_cbor"])
            if not issubclass(cls, WingRidersV2SSPState):
                raise NotAPoolError("Invalid Datum")
        except TypeError:
            if issubclass(cls, WingRidersV2SSPState):
                raise NotAPoolError("Invalid Datum")
            datum = WingRidersV2PoolDatum.from_cbor(values["datum_cbor"])

        values["fee"] = int(
            (
                datum.swap_fee_in_basis
                + datum.protocol_fee_in_basis
                + datum.project_fee_in_basis
            )
            * 10000
            / datum.fee_basis,
        )

        return datum

    @classmethod
    def post_init(cls, values):
        datum = cls._post_init(values)

        assets = values["assets"]

        if len(assets) == 2:
            assets.root[assets.unit(0)] -= 3000000

        assets.root[assets.unit(0)] -= (
            datum.treasury_a + datum.project_treasury_a + datum.reserve_treasury_a
        )
        assets.root[assets.unit(1)] -= (
            datum.treasury_b + datum.project_treasury_b + datum.reserve_treasury_b
        )

    def deposit(
        self,
        in_assets: Assets | None = None,
        out_assets: Assets | None = None,
    ):
        merged_assets = in_assets + out_assets
        if "lovelace" in merged_assets:
            return Assets(lovelace=4000000) - self.batcher_fee(
                in_assets=in_assets,
                out_assets=out_assets,
            )
        else:
            return self._deposit

    def batcher_fee(
        self,
        in_assets: Assets | None = None,
        out_assets: Assets | None = None,
        extra_assets: Assets | None = None,
    ):
        merged_assets = in_assets + out_assets
        if "lovelace" in merged_assets:
            if merged_assets["lovelace"] <= 250000000:
                return Assets(lovelace=850000)
            elif merged_assets["lovelace"] <= 500000000:
                return Assets(lovelace=1500000)
        return Assets(lovelace=2000000)


class WingRidersV2SSPState(AbstractStableSwapPoolState, WingRidersV2CPPState):
    """WingRiders SSP state."""

    fee: int = 6
    _batcher = Assets(lovelace=2000000)
    _deposit = Assets(lovelace=2000000)
    _stake_address = Address.from_primitive(
        "addr1wy3ksr4xwqd4dukp9tnemzheflfk75ym0vq8q2w8ecg5ssqmfdjaz",
    )
    _pool_datum_parsed: WingRidersV2PoolDatum | None = None

    asset_mulitipliers: ClassVar[list[int]] = [1, 1]

    @classmethod
    def pool_selector(cls) -> PoolSelector:
        return PoolSelector(
            addresses=["addr1wx2x4c3ggv8jl3j24ze6ewgsacn7nvly0250jf06cfurfggd7zqtl"],
            assets=cls.dex_policy(),
        )

    @classmethod
    def pool_datum_class(cls) -> type[WingRidersV2SSPoolDatum]:
        return WingRidersV2SSPoolDatum

    @classmethod
    def post_init(cls, values):
        super().post_init(values)

        datum: WingRidersV2SSPoolDatum = WingRidersV2SSPoolDatum.from_cbor(
            values["datum_cbor"],
        )

        cls.asset_mulitipliers = [
            datum.pool_specifics.scale_a,
            datum.pool_specifics.scale_b,
        ]

    def swap_datum(
        self,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        extra_assets: Assets | None = None,
        address_target: Address | None = None,
        datum_target: PlutusData | None = None,
    ) -> PlutusData:
        return self.order_datum_class().create_datum(
            address_source=address_source,
            in_assets=in_assets,
            out_assets=out_assets,
            batcher_fee=self.batcher_fee(
                in_assets=in_assets,
                out_assets=out_assets,
                extra_assets=extra_assets,
            ),
            deposit=self.deposit(in_assets=in_assets, out_assets=out_assets),
            address_target=address_target,
            datum_target=datum_target,
            a_scale=self.asset_mulitipliers[0],
            b_scale=self.asset_mulitipliers[1],
        )



"""SundaeSwap DEX module."""
import time
from dataclasses import dataclass
from functools import lru_cache
from typing import Any
from typing import ClassVar
from typing import List
from typing import Union

from pycardano import Address
from pycardano import Datum
from pycardano import PlutusData
from pycardano import PlutusV1Script
from pycardano import PlutusV2Script
from pycardano import VerificationKeyHash

from charli3_dendrite.backend import get_backend
from charli3_dendrite.dataclasses.datums import AssetClass
from charli3_dendrite.dataclasses.datums import OrderDatum
from charli3_dendrite.dataclasses.datums import PlutusFullAddress
from charli3_dendrite.dataclasses.datums import PlutusNone
from charli3_dendrite.dataclasses.datums import PlutusPartAddress
from charli3_dendrite.dataclasses.datums import PlutusScriptAddress
from charli3_dendrite.dataclasses.datums import PoolDatum
from charli3_dendrite.dataclasses.datums import ReceiverDatum
from charli3_dendrite.dataclasses.models import Assets
from charli3_dendrite.dataclasses.models import OrderType
from charli3_dendrite.dataclasses.models import PoolSelector
from charli3_dendrite.dexs.amm.amm_types import AbstractConstantProductPoolState
from charli3_dendrite.dexs.core.errors import InvalidPoolError
from charli3_dendrite.dexs.core.errors import NoAssetsError
from charli3_dendrite.dexs.core.errors import NotAPoolError


@dataclass
class AtoB(PlutusData):
    """A to B swap direction."""

    CONSTR_ID = 0


@dataclass
class BtoA(PlutusData):
    """B to A swap direction."""

    CONSTR_ID = 1


@dataclass
class AmountOut(PlutusData):
    """Minimum amount to receive."""

    CONSTR_ID = 0
    min_receive: int


@dataclass
class SwapConfig(PlutusData):
    """Swap configuration."""

    CONSTR_ID = 0

    direction: Union[AtoB, BtoA]
    amount_in: int
    amount_out: AmountOut


@dataclass
class DepositPairQuantity(PlutusData):
    """Deposit pair quantity."""

    CONSTR_ID = 0
    amount_a: int
    amount_b: int


@dataclass
class DepositPair(PlutusData):
    """Deposit pair."""

    CONSTR_ID = 1
    assets: DepositPairQuantity


@dataclass
class DepositConfig(PlutusData):
    """Deposit configuration."""

    CONSTR_ID = 2

    deposit_pair: DepositPair


@dataclass
class WithdrawConfig(PlutusData):
    """Withdraw configuration."""

    CONSTR_ID = 1

    amount_lp: int


@dataclass
class SundaeV3PlutusNone(PlutusData):
    CONSTR_ID = 0


@dataclass
class SundaeV3ReceiverDatumHash(PlutusData):
    CONSTR_ID = 1

    datum_hash: bytes


@dataclass
class SundaeV3ReceiverInlineDatum(PlutusData):
    CONSTR_ID = 2

    datum: Datum


@dataclass
class SundaeAddressWithDatum(PlutusData):
    """SundaeSwap address with datum."""

    CONSTR_ID = 0

    address: Union[PlutusFullAddress, PlutusScriptAddress]
    datum: Union[
        ReceiverDatum,
        PlutusNone,
    ]

    @classmethod
    def from_address(cls, address: Address) -> "SundaeAddressWithDatum":
        """Create a new address with datum."""
        return cls(address=PlutusFullAddress.from_address(address), datum=PlutusNone())


@dataclass
class SundaeV3AddressWithDatum(PlutusData):
    CONSTR_ID = 0

    address: Union[PlutusFullAddress, PlutusScriptAddress]
    datum: Union[
        SundaeV3PlutusNone,
        SundaeV3ReceiverDatumHash,
        SundaeV3ReceiverInlineDatum,
    ]

    @classmethod
    def from_address(cls, address: Address):
        return cls(
            address=PlutusFullAddress.from_address(address),
            datum=SundaeV3PlutusNone(),
        )


@dataclass
class SundaeAddressWithDestination(PlutusData):
    """For now, destination is set to none, should be updated."""

    CONSTR_ID = 0

    address: SundaeAddressWithDatum
    destination: Union[PlutusPartAddress, PlutusNone]

    @classmethod
    def from_address(cls, address: Address) -> "SundaeAddressWithDestination":
        """Create a new address with destination."""
        null = SundaeAddressWithDatum.from_address(address)
        return cls(address=null, destination=PlutusNone())


@dataclass
class SundaeOrderDatum(OrderDatum):
    """SundaeSwap order datum."""

    ident: bytes
    address: SundaeAddressWithDestination
    fee: int
    swap: Union[DepositConfig, SwapConfig, WithdrawConfig]

    @classmethod
    def create_datum(
        cls,
        ident: bytes,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        fee: int,
    ) -> "SundaeOrderDatum":
        """Create a new order datum."""
        full_address = SundaeAddressWithDestination.from_address(address_source)
        merged = in_assets + out_assets
        if in_assets.unit() == merged.unit():
            direction = AtoB()
        else:
            direction = BtoA()
        swap = SwapConfig(
            direction=direction,
            amount_in=in_assets.quantity(),
            amount_out=AmountOut(min_receive=out_assets.quantity()),
        )

        return cls(ident=ident, address=full_address, fee=fee, swap=swap)

    def address_source(self) -> Address:
        """Get the source address."""
        return self.address.address.address.to_address()

    def requested_amount(self) -> Assets:
        """Get the requested amount."""
        if isinstance(self.swap, SwapConfig):
            if isinstance(self.swap.direction, AtoB):
                return Assets({"asset_b": self.swap.amount_out.min_receive})
            else:
                return Assets({"asset_a": self.swap.amount_out.min_receive})
        else:
            return Assets({})

    def order_type(self) -> OrderType | None:
        """Get the order type."""
        order_type = None
        if isinstance(self.swap, SwapConfig):
            order_type = OrderType.swap
        elif isinstance(self.swap, DepositConfig):
            order_type = OrderType.deposit
        elif isinstance(self.swap, WithdrawConfig):
            order_type = OrderType.withdraw

        return order_type


@dataclass
class SwapV3Config(PlutusData):
    CONSTR_ID = 1
    in_value: List[Union[int, bytes]]
    out_value: List[Union[int, bytes]]


@dataclass
class DepositV3Config(PlutusData):
    CONSTR_ID = 2
    values: List[List[Union[int, bytes]]]


@dataclass
class WithdrawV3Config(PlutusData):
    CONSTR_ID = 3
    in_value: List[Union[int, bytes]]


# @dataclass
# class ZapInV3Config(PlutusData):
#     CONSTR_ID = 4
#     in_value: List[Union[int, bytes]]
#     out_value: List[Union[int, bytes]]


# @dataclass
# class ZapOutV3Config(PlutusData):
#     CONSTR_ID = 5
#     token_a: int
#     token_b: int


@dataclass
class DonateV3Config(PlutusData):
    CONSTR_ID = 4
    in_value: List[Union[int, bytes]]
    out_value: List[Union[int, bytes]]


@dataclass
class Ident(PlutusData):
    CONSTR_ID = 0
    payload: bytes


@dataclass
class SundaeV3OrderDatum(OrderDatum):
    CONSTR_ID = 0

    ident: Ident
    owner: PlutusPartAddress
    max_protocol_fee: int
    destination: SundaeV3AddressWithDatum
    swap: Union[
        DepositV3Config,
        WithdrawV3Config,
        # ZapInV3Config,
        # ZapOutV3Config,
        DonateV3Config,
        SwapV3Config,
    ]
    extension: Union[Datum, bytes]

    @classmethod
    def create_datum(
        cls,
        ident: bytes,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        fee: int,
    ):
        full_address = SundaeV3AddressWithDatum.from_address(address_source)
        merged = in_assets + out_assets
        if in_assets.unit() == merged.unit():
            direction = AtoB()
        else:
            direction = BtoA()
        swap = SwapConfig(
            direction=direction,
            amount_in=in_assets.quantity(),
            amount_out=AmountOut(min_receive=out_assets.quantity()),
        )

        if in_assets.unit() == "lovelace":
            in_policy = in_name = ""
        else:
            in_policy = in_assets.unit()[:56]
            in_name = in_assets.unit()[56:]

        if out_assets.unit() == "lovelace":
            out_policy = out_name = ""
        else:
            out_policy = out_assets.unit()[:56]
            out_name = out_assets.unit()[56:]

        in_value = [
            bytes.fromhex(in_policy),
            bytes.fromhex(in_name),
            in_assets.quantity(),
        ]
        out_value = [
            bytes.fromhex(out_policy),
            bytes.fromhex(out_name),
            out_assets.quantity(),
        ]

        return cls(
            ident=Ident(payload=ident),
            owner=PlutusPartAddress(address=address_source.staking_part.payload),
            max_protocol_fee=fee,
            destination=full_address,
            swap=SwapV3Config(in_value=in_value, out_value=out_value),
            extension=PlutusData().to_cbor(),
        )

    def address_source(self) -> Address:
        return Address(staking_part=VerificationKeyHash(self.owner.address))

    def requested_amount(self) -> Assets:
        if isinstance(self.swap, SwapV3Config):
            asset = (self.swap.out_value[0] + self.swap.out_value[1]).hex()
            if asset == "":
                asset = "lovelace"
            return Assets({asset: self.swap.out_value[2]})
        else:
            return Assets({})

    def order_type(self) -> OrderType:
        order_type = None
        if isinstance(self.swap, SwapV3Config):
            order_type = OrderType.swap
        elif isinstance(self.swap, DepositV3Config):
            order_type = OrderType.deposit
        elif isinstance(self.swap, WithdrawV3Config):
            order_type = OrderType.withdraw

        return order_type


@dataclass
class LPFee(PlutusData):
    """Liquidity pool fee."""

    CONSTR_ID = 0
    numerator: int
    denominator: int


@dataclass
class LiquidityPoolAssets(PlutusData):
    """Liquidity pool assets."""

    CONSTR_ID = 0
    asset_a: AssetClass
    asset_b: AssetClass


@dataclass
class SundaePoolDatum(PoolDatum):
    """SundaeSwap pool datum."""

    assets: LiquidityPoolAssets
    ident: bytes
    last_swap: int
    fee: LPFee

    def pool_pair(self) -> Assets | None:
        return self.assets.asset_a.assets + self.assets.asset_b.assets


@dataclass
class SundaeV3PoolDatum(PlutusData):
    CONSTR_ID = 0
    ident: bytes
    assets: List[List[bytes]]
    circulation_lp: int
    bid_fees_per_10_thousand: int
    ask_fees_per_10_thousand: int
    fee_manager: Union[PlutusNone, Any]
    market_open: int  # time in milliseconds
    protocol_fees: int

    def pool_pair(self) -> Assets | None:
        assets = {}
        for asset in self.assets:
            assets[asset[0].hex() + asset[1].hex()] = 0
        if "" in assets:
            assets.pop("")
            assets["lovelace"] = 0
        return Assets(**assets)


@dataclass
class SundaeV3Settings(PlutusData):
    CONSTR_ID = 0
    settings_admin: Datum  # NativeScript
    metadata_admin: PlutusFullAddress
    treasury_admin: Datum  # NativeScript
    treasury_address: PlutusFullAddress
    treasury_allowance: List[int]
    authorized_scoopers: Union[PlutusNone, Any]  # List[PlutusPartAddress]]
    authorized_staking_keys: List[Any]
    base_fee: int
    simple_fee: int
    strategy_fee: int
    pool_creation_fee: int
    extensions: Union[Datum, bytes]


class SundaeSwapCPPState(AbstractConstantProductPoolState):
    """SundaeSwap constant product pool state."""

    fee: int = 0
    _batcher = Assets(lovelace=2500000)
    _deposit = Assets(lovelace=2000000)
    _stake_address: ClassVar[Address] = Address.from_primitive(
        "addr1wxaptpmxcxawvr3pzlhgnpmzz3ql43n2tc8mn3av5kx0yzs09tqh8",
    )

    @classmethod
    def dex(cls) -> str:
        """Get the DEX name."""
        return "SundaeSwap"

    @classmethod
    def order_selector(self) -> list[str]:
        """Get the order selector."""
        return [self._stake_address.encode()]

    @classmethod
    def pool_selector(cls) -> PoolSelector:
        return PoolSelector(
            addresses=["addr1w9qzpelu9hn45pefc0xr4ac4kdxeswq7pndul2vuj59u8tqaxdznu"],
        )

    @property
    def swap_forward(self) -> bool:
        """Check if swap forwarding is enabled."""
        return False

    @property
    def stake_address(self) -> Address:
        """Get the stake address."""
        return self._stake_address

    @classmethod
    def order_datum_class(self) -> type[SundaeOrderDatum]:
        """Get the order datum class."""
        return SundaeOrderDatum

    @classmethod
    def pool_datum_class(cls) -> type[SundaePoolDatum]:
        """Get the pool datum class."""
        return SundaePoolDatum

    @property
    def pool_id(self) -> str:
        """A unique identifier for the pool."""
        return self.pool_nft.unit()

    @classmethod
    def skip_init(cls, values) -> bool:
        """Skip the initialization process."""
        if "pool_nft" in values and "dex_nft" in values and "fee" in values:
            try:
                super().extract_pool_nft(values)
            except InvalidPoolError:
                raise NotAPoolError("No pool NFT found.")
            if len(values["assets"]) == 3:
                # Send the ADA token to the end
                if isinstance(values["assets"], Assets):
                    values["assets"].root["lovelace"] = values["assets"].root.pop(
                        "lovelace",
                    )
                else:
                    values["assets"]["lovelace"] = values["assets"].pop("lovelace")
            values["assets"] = Assets.model_validate(values["assets"])
            return True
        else:
            return False

    @classmethod
    def extract_pool_nft(cls, values) -> Assets:
        """Extract the pool NFT."""
        try:
            super().extract_pool_nft(values)
        except InvalidPoolError:
            if len(values["assets"]) == 0:
                raise NoAssetsError
            else:
                raise NotAPoolError("No pool NFT found.")

    @classmethod
    def pool_policy(cls) -> list[str]:
        """Get the pool policy."""
        return ["0029cb7c88c7567b63d1a512c0ed626aa169688ec980730c0473b91370"]

    @classmethod
    def post_init(cls, values):
        """Post initialization."""
        super().post_init(values)

        assets = values["assets"]
        datum = SundaePoolDatum.from_cbor(values["datum_cbor"])

        if len(assets) == 2:
            assets.root[assets.unit(0)] -= 2000000
        elif len(assets) == 1:
            msg = "Reserves contains only 1 token."
            raise InvalidPoolError(msg)

        numerator = datum.fee.numerator
        denominator = datum.fee.denominator
        values["fee"] = int(numerator * 10000 / denominator)

    def swap_datum(
        self,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        extra_assets: Assets | None = None,
        address_target: Address | None = None,
        datum_target: PlutusData | None = None,
    ) -> PlutusData:
        """Create a swap datum."""
        if self.swap_forward and address_target is not None:
            print(f"{self.__class__.__name__} does not support swap forwarding.")

        ident = bytes.fromhex(self.pool_nft.unit()[60:])

        return SundaeOrderDatum.create_datum(
            ident=ident,
            address_source=address_source,
            in_assets=in_assets,
            out_assets=out_assets,
            fee=self.batcher_fee(in_assets=in_assets, out_assets=out_assets).quantity(),
        )


class SundaeSwapV3CPPState(AbstractConstantProductPoolState):
    fee: int | list[int] = [30, 30]
    _batcher = Assets(lovelace=1280000)
    _deposit = Assets(lovelace=2000000)
    _stake_address: ClassVar[Address] = Address.from_primitive(
        "addr1z8ax5k9mutg07p2ngscu3chsauktmstq92z9de938j8nqa7zcka2k2tsgmuedt4xl2j5awftvqzmmv3vs2yduzqxfcmsyun6n3",
    )

    def batcher_fee(
        self,
        in_assets: Assets | None = None,
        out_assets: Assets | None = None,
        extra_assets: Assets | None = None,
    ) -> Assets:
        return self.__class__._batcher_fee
        # return Assets(lovelace=1280000)

    @classmethod
    def dex(cls) -> str:
        return "SundaeSwapV3"

    @classmethod
    def default_script_class(self) -> type[PlutusV1Script] | type[PlutusV2Script]:
        return PlutusV2Script

    @classmethod
    def order_selector(self) -> list[str]:
        return [self._stake_address.encode()]

    @classmethod
    def pool_selector(cls) -> PoolSelector:
        return PoolSelector(
            addresses=["addr1w8srqftqemf0mjlukfszd97ljuxdp44r372txfcr75wrz2cp9h6f8"],
        )

    @classmethod
    def pool_policy(cls) -> list[str]:
        return ["e0302560ced2fdcbfcb2602697df970cd0d6a38f94b32703f51c312b"]

    @property
    def swap_forward(self) -> bool:
        return False

    @property
    def stake_address(self) -> Address:
        return self._stake_address

    @classmethod
    def order_datum_class(self) -> type[SundaeV3OrderDatum]:
        return SundaeV3OrderDatum

    @classmethod
    def pool_datum_class(self) -> type[SundaeV3PoolDatum]:
        return SundaeV3PoolDatum

    @property
    def pool_id(self) -> str:
        """A unique identifier for the pool."""
        return self.pool_nft.unit()

    @classmethod
    @lru_cache
    def get_batcher_fee(cls, last_check: int = time.time() // 3600):
        try:
            settings = get_backend().get_datum_from_address(
                address=Address.decode(
                    "addr1w9ke67k2ckdyg60v22ajqugxze79e0ax3yqgl7nway4vc5q84hpqs",
                ),
                asset="6d9d7acac59a4469ec52bb207106167c5cbfa689008ffa6ee92acc5073657474696e6773",
            )
        except ValueError:
            print(
                "No settings found for SundaeSwapV3, likely because no backend is set.",
            )
            return

        datum = SundaeV3Settings.from_cbor(settings.datum_cbor)
        cls._batcher_fee = Assets(lovelace=datum.simple_fee + datum.base_fee)

    @classmethod
    def skip_init(cls, values) -> bool:
        if "pool_nft" in values:
            try:
                super().extract_pool_nft(values)
            except InvalidPoolError:
                raise NotAPoolError("No pool NFT found.")
            if len(values["assets"]) == 3:
                # Send the ADA token to the end
                if isinstance(values["assets"], Assets):
                    values["assets"].root["lovelace"] = values["assets"].root.pop(
                        "lovelace",
                    )
                else:
                    values["assets"]["lovelace"] = values["assets"].pop("lovelace")

            datum = SundaeV3PoolDatum.from_cbor(values["datum_cbor"])
            values["fee"] = datum.bid_fees_per_10_thousand
            values["assets"] = Assets.model_validate(values["assets"])

            cls.get_batcher_fee()
            return True
        else:
            return False

    @classmethod
    def extract_pool_nft(cls, values) -> Assets:
        try:
            super().extract_pool_nft(values)
        except InvalidPoolError:
            if len(values["assets"]) == 0:
                raise NoAssetsError
            else:
                raise NotAPoolError("No pool NFT found.")

    @classmethod
    def post_init(cls, values):
        super().post_init(values)

        assets = values["assets"]
        datum = SundaeV3PoolDatum.from_cbor(values["datum_cbor"])

        if len(assets) == 2:
            assets.root[assets.unit(0)] -= datum.protocol_fees

        values["fee"] = [datum.bid_fees_per_10_thousand, datum.ask_fees_per_10_thousand]

        cls.get_batcher_fee()

    def swap_datum(
        self,
        address_source: Address,
        in_assets: Assets,
        out_assets: Assets,
        extra_assets: Assets | None = None,
        address_target: Address | None = None,
        datum_target: PlutusData | None = None,
    ) -> PlutusData:
        if self.swap_forward and address_target is not None:
            print(f"{self.__class__.__name__} does not support swap forwarding.")

        ident = bytes.fromhex(self.pool_nft.unit()[64:])

        datum = SundaeV3OrderDatum.create_datum(
            ident=ident,
            address_source=address_source,
            in_assets=in_assets,
            out_assets=out_assets,
            fee=self.batcher_fee(in_assets=in_assets, out_assets=out_assets).quantity(),
        )

        return datum
